-- Ŀ
--  Save-AVI.E V3.02 (C) 1999 Legend Interactive Ltd            AVI Library 
-- 
-- This library contains functions to encode images or screens into
-- Windows AVI raw 8/16/24-Bit RGB format.

-- 8-Bit should be playble on:-
--  Windows Media Player (mplay.exe)
--  Video for Windows (videdit.exe)

-- 24-Bit is not meant to be played as is, but to be compressed using one of
-- the many codecs around. I hope to include some codecs in a while.

-- 16-Bit is half way between 8 and 24 bit

-- This library does not include sound support, but I hope to include
-- it in a later version.

-- Please visit our web site for more updates - www.legend-i.demon.co.uk

include file.e -- RDS Lib
include image.e -- RDS Lib
include core-avi.e
include iprocess.e

AviFramesPerSecond = 15 -- Try a sensible value (1 ~ 50)
AviBitsPerPixel    = 24 -- Try 8, 16 or 24
AviHeader = repeat(0,4096) -- Don't touch !
AviIndex = 0

procedure hdrb(integer byt) -- Push byte into header
  AviPtr1 = AviPtr1 + 1
  AviHeader[AviPtr1] = and_bits(byt,#FF)
end procedure

procedure hdrs(sequence str) -- Push String to header
  for i = 1 to length(str) do
    hdrb(str[i])
  end for
end procedure

procedure hdri(integer int) -- Push Integer to header
  hdrb(int)
  int = floor(int / 256)
  hdrb(int)
end procedure

procedure hdrw(integer int) -- Push Word to header
  hdrb(int)
  int = floor(int / 256)
  hdrb(int)
  int = floor(int / 256)
  hdrb(int)
  int = floor(int / 256)
  hdrb(int)
end procedure

-- Usage :-
-- OutputFileHandle = SaveHeaderAVI(OutputFilename,Width,Height,Frames)
global function SaveHeaderAVI
  (sequence ofn,integer xsize,integer ysize,integer frames)
  integer isize,msize,PlayRate,FrameDelay,DataRate,FileLen
  atom    ofh
  sequence pal
  
  AviPtr1 = 0
  AviFrameOffset = {}
  AviWidth = xsize
  AviHeight = ysize
  
  isize = xsize * ysize * floor(AviBitsPerPixel / 8)
  msize = ((isize + 8) * frames)
  FileLen = 0 -- Won't know this value until we finish
  
  AviFrameSize = isize
  
  AviTimeScale = 1000000
  PlayRate = AviTimeScale * AviFramesPerSecond
  FrameDelay = floor(AviTimeScale / AviFramesPerSecond)
  DataRate = AviFramesPerSecond * isize
  
  AviWidth = xsize
  AviHeight = ysize
  
  -- Fill in header details
  hdrs("RIFF")
  hdrw(0)
  hdrs("AVI LIST")
  hdrw(1208)
  hdrs("hdrlavih")
  hdrw(#38)
  hdrw(FrameDelay)
  hdrw(DataRate)
  hdrw(0) -- {Reserved}
  if AviIndex > 0 then
    hdrw(#10) -- Index = yes {dwFlags}
  else
    hdrw(0) -- Index = no {dwFlags}
  end if
  hdrw(frames)
  hdrw(0) -- {Initial frame}
  hdrw(1) -- Video only {Streams}
  hdrw(0) -- {Suggested buf size}
  hdrw(xsize)
  hdrw(ysize)
  hdrw(AviTimeScale)
  hdrw(PlayRate)
  hdrw(0) -- {Starting time}
  hdrw(0) -- {AVI data CHUNK size}
  
  hdrs("LIST")
  hdrw(1132)
  hdrs("strlstrh")
  hdrw(#30)
  hdrs("vids")
  hdrs("RGB ")
  hdrw(0)
  hdrw(0)
  hdrw(0)
  hdrw(FrameDelay)
  hdrw(AviTimeScale)
  hdrw(0)
  hdrw(frames)
  hdrw(AviFrameSize)
  hdrw(0)
  hdrw(0)
  hdrs("strf")
  hdrw(1064)
  hdrw(40)
  hdrw(xsize)
  hdrw(ysize)
  hdri(1) -- Planes
  hdri(AviBitsPerPixel)
  hdrw(0) -- Compression type 0=RGB, 1=RLE8, 2=RLE4
  hdrw(AviFrameSize)
  hdrw(0) -- x pels / metre
  hdrw(0) -- y pels / metre
  hdrw(0) -- colours used
  hdrw(0) -- colours important
  if AviBitsPerPixel < 9 then -- Grab current palette
    if length(AviPalette) < 1 then -- Grab palette from screen
      AviPalette = (get_all_palette() * 4)
    end if
    for p = 1 to 256 do -- Need 256 * 4 of space
      if p <= length(AviPalette) then -- Fill with palette
	hdrb(AviPalette[p][3])
	hdrb(AviPalette[p][2])
	hdrb(AviPalette[p][1])
	hdrb(0)
      else -- Fill with 0's
	hdrw(0)
      end if
    end for
  else -- No palette, just step over it.
    AviPtr1 = AviPtr1 + 1024
  end if
  
  hdrs("junk")
  hdrw(2848)
  
  AviPtr1 = #00FF4
  hdrs("LIST")
  hdrw(msize)
  hdrs("movi")
  
  ofh = open(ofn,"wb")
  if ofh < 1 then -- File error !
    return ofh    
  end if
  
  puts(ofh,AviHeader)
  AviFrame1 = where(ofh)
  
  return ofh
end function

-- This will take a 2D sequence containing RGB pixels and write it
-- as an AVI frame.
-- Usage :-
-- SaveFrameAVI(OutputFileHandle, ImageSequence)
global procedure SaveFrameAVI(integer ofh,sequence Im)
  integer xsize,ysize,isize,r,g,b,pix
  
  if not sequence(Im) then
    puts(1,"ERROR:Bad image type %d\n")
    close(ofh)
    abort(1)
  end if
  
  xsize = length(Im[1])
  ysize = length(Im)
  if (xsize != AviWidth) or (ysize != AviHeight) then
    printf(1,"ERROR: All images must be same size \n",{})
    printf(1,"First image was %d x %d\n",{AviWidth,AviHeight})
    printf(1,"This image was  %d x %d\n",{xsize,ysize})
    close(ofh)
    abort(1)
  end if
  
  isize = xsize * ysize 
  if AviBitsPerPixel = 24 then
    AviFrameOffset = AviFrameOffset & where(ofh) - AviFrame1
    puts(ofh,"00db")
    WriteLongI86(ofh,isize * 3)
    for y = ysize to 1 by -1 do
      for x = 1 to xsize  do
	pix = Im[y][x] -- Get RGB24 Pixel
	r = and_bits(pix,#FF)
	g = and_bits((pix / 256),#FF)
	b = and_bits((pix / 65536),#FF)
	puts(ofh,{b,g,r})
      end for
    end for
  elsif AviBitsPerPixel = 16 then
    AviFrameOffset = AviFrameOffset & where(ofh) - AviFrame1
    puts(ofh,"00db")
    WriteLongI86(ofh,isize * 2)
    for y = ysize to 1 by -1 do
      for x = 1 to xsize do
	pix = Im[y][x] -- Get RGB24 pixel
	r = floor(and_bits(pix,#FF) / 8)
	g = floor(and_bits((pix / 256),#FF) / 8)
	b = floor(and_bits((pix / 65536),#FF) / 8)
	WriteIntI86(ofh,(b + (g * 32) + (r * 1024)))
      end for
    end for
  elsif AviBitsPerPixel = 8 then
    puts(ofh,"00db")
    WriteLongI86(ofh,isize)
    AviFrameOffset = AviFrameOffset & where(ofh) - AviFrame1
    for y = ysize to 1 by -1 do
      puts(ofh,Im[y])
    end for
  else
    puts(1,"Unsupported format !\n")
    close(ofh)
    abort(1)
  end if
end procedure

-- This is used to finish the AVI sequence
-- Usage :-
-- CloseHeaderAVI(OutputFileHandle)
global procedure CloseHeaderAVI(integer ofh)
  integer max,ok,FileLen
  
  if AviIndex > 0 then -- I'm not sure if this has any useful purpose !!
    max = length(AviFrameOffset)
    puts(ofh,"idx1")
    WriteLongI86(ofh,max * 16)
    for cnt = 1 to max do
      puts(ofh,"00db")
      WriteLongI86(ofh,1)
      WriteLongI86(ofh,AviFrameOffset[cnt])
      WriteLongI86(ofh,AviFrameSize)
    end for
  end if
  AviFrameOffset = {}
  
  FileLen = where(ofh) -- Now we can determine the file size ...
  ok = seek(ofh,0)
  AviPtr1 = 4
  --hdrs("RIFF")
  hdrw(FileLen)
  puts(ofh,AviHeader) -- ... and write into the header
  close (ofh)
end procedure


-- This will take a copy of the current graphics screen and write
-- it out as a 8-Bit AVI frame.
-- Ensure you have set AviBitsPerPixel = 8 for best performance
-- Usage :-
-- SaveScreenAVI(OutputFileHandle)
global procedure SaveScreenAVI(integer ofh)
  integer xsize,ysize
  sequence vc,image
  
  vc = video_config()
  xsize = vc[5]
  ysize = vc[6]
  
  if (xsize != AviWidth) or (ysize != AviHeight) then
    printf(1,"ERROR: All images must be same size \n",{})
    printf(1,"First image was %d x %d\n",{AviWidth,AviHeight})
    printf(1,"This image was  %d x %d\n",{xsize,ysize})
    close(ofh)
    abort(1)
  end if
  
  image = save_image({0,0},{xsize - 1,ysize - 1})
  if AviBitsPerPixel > 15 then
    image = PAL2RGB24(image,{}) -- Convert to 24 bit True-Colour
  end if
  SaveFrameAVI(ofh,image)
  
end procedure

